*=======================================================================
* WCSLIB 8.8 - an implementation of the FITS WCS standard.
* Copyright (C) 1995-2026, Mark Calabretta
*
* This file is part of WCSLIB.
*
* WCSLIB is free software: you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as published
* by the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* WCSLIB is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
* License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with WCSLIB.  If not, see http://www.gnu.org/licenses.
*
* Author: Mark Calabretta, Australia Telescope National Facility, CSIRO.
* http://www.atnf.csiro.au/computing/software/wcs
* $Id: log_data.f,v 8.8 2026/05/22 11:19:30 mcalabre Exp $
*=======================================================================

      BLOCK DATA LOG_BLOCK_DATA

      CHARACTER LOG_ERRMSG(0:4)*80

      COMMON /LOG_DATA/ LOG_ERRMSG

      DATA LOG_ERRMSG /
     :  'Success',
     :  ' ',
     :  'Invalid log-coordinate reference value',
     :  'One or more of the x coordinates were invalid',
     :  'One or more of the world coordinates were invalid'/

      END
